/* CallableDesc.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.util.*;

/**
 * This is the superclass for method and constructor descriptions.
 */
public abstract class CallableDesc extends MemberDesc implements Comparable {
	/**
	 * A list of parameters (<tt>Parameter</tt>) that this
	 * callable takes.
	 */ 
	protected List parameters = new LinkedList();
	/**
	 * A list of exception (<tt>String</tt>) thrown by this
	 * callable.
	 */
	protected List exceptions = new LinkedList();

	/**
	 * A parameter for the method/constructor.
	 */
	protected class Parameter implements Comparable {
		/** Name of parameter. */
		public String name;
		/** Type of parameter, for example <tt>float</tt>. */
		public String type;

		/** Create a new parameter specifying name and type. */
		public Parameter(String name, String type) {
			this.name = name;
			this.type = type;
		}

		public int compareTo(Object o) {
			Parameter p = (Parameter) o;
			return this.type.compareTo(p.type);
		}
	}

	/**
	 * Create a <tt>CallableDesc</tt> object with the
	 * specified name.
	 */
	protected CallableDesc(ClassDesc classDesc, String name) {
		super(classDesc, name);
	}

	/**
	 * Add a parameter.
	 */
	public void addParameter(String name, String type) {
		parameters.add(new Parameter(name, type));
	}

	/**
	 * Add an exception.
	 */
	public void addException(String name) {
		exceptions.add(name);
	}

	public int compareTo(Object o) {
		int superDiff = super.compareTo(o);
		if (superDiff != 0) {
			return superDiff;
		}

		CallableDesc c = (CallableDesc) o;

		/* Compare each parameter. */
		Iterator it1 = this.parameters.iterator();
		Iterator it2 = c.parameters.iterator();
		while (it1.hasNext() && it2.hasNext()) {
			Parameter p1 = (Parameter) it1.next();
			int diff = p1.compareTo(it2.next());
			if (diff != 0) {
				return diff;
			}
		}
		if (it1.hasNext() ^ it2.hasNext()) {
			return (it1.hasNext() ? -1 : 1);
		}

		return 0;
	}

	/**
	 * Generate java source for this method/constructor.
	 */
	protected void generate(JavaSourceWriter out, String returnType,
			int bodyLevel) {
		Iterator it;
		boolean first;

		/* Print return type, modifiers and name. */
		out.print(modifiers+" ");
		if (returnType != null) {
			out.print(returnType+" ");
		}
		out.print(name+"(");

		/* Print parameters. */
		first = true;
		it = parameters.iterator();
		while (it.hasNext()) {
			Parameter param = (Parameter) it.next();
			if (first) {
				first = false;
			} else {
				out.print(", ");
			}
			out.print(param.type+" "+param.name);
		}
		out.print(")");

		/* Print exceptions. */
		first = true;
		it = exceptions.iterator();
		while (it.hasNext()) {
			String ex = (String) it.next();
			if (first) {
				out.println();
				out.print("throws ");
				first = false;
			} else {
				out.print(", ");
			}
			out.print(ex);
		}

		/* Print body. */
		if (hasModifier("abstract")
				|| classDesc.hasModifier("interface")
				|| bodyLevel == 0) {
			out.println(";");
		} else if (bodyLevel == 1) {
			out.println(" {");
			out.println("}");
		} else if (bodyLevel == 2) {
			out.println(" {");
			out.increaseLevel();
			String value = MemberDesc.getTypeDefaultValue(returnType);
			if (value != null) {
				out.println("return "+value+";");
			}
			out.decreaseLevel();
			out.println("}");
		} else if (bodyLevel == 3) {
			out.println(" {");
			out.increaseLevel();
			out.println("throw new UnsupportedOperationException(\"unimplemented\");");
			out.decreaseLevel();
			out.println("}");
		}
	}
}
