/* ClassDesc.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.*;

/**
 * An object of this class represents the information
 * of a class.
 */
public class ClassDesc extends EntityDesc implements Comparable {
	private String classPackage;
	private String className;
	private String parentPackage;
	private String parentName;
	private List implems = new LinkedList();
	private List fields = new LinkedList();
	private List constructors = new LinkedList();
	private List methods = new LinkedList();

	/**
	 * Construct a class description without a name.
	 */
	public ClassDesc() {
	}

	/**
	 * Return the set of fields in this class.
	 */
	public Collection getFields() {
		return fields;
	}

	/**
	 * Set the class modifiers.
	 */
	public void setModifiers(String modifiers) {
		this.modifiers = modifiers;
	}

	/**
	 * Add a new field to this class.
	 */
	public void addField(FieldDesc field) {
		fields.add(field);
	}

	/**
	 * Add a new method to this class.
	 */
	public void addMethod(MethodDesc method) {
		methods.add(method);
	}

	/**
	 * Add a new constructor to this class.
	 */
	public void addConstructor(ConstructorDesc constructor) {
		constructors.add(constructor);
	}

	/**
	 * Add an implementated interface.
	 */
	public void addImplements(String name) {
		implems.add(name);
	}

	/**
	 * Set the name of this class.
	 */
	public void setName(String name) {
		this.className = name;
	}

	/**
	 * Set the full name (package and class name) of this class.
	 */
	public void setFullName(String name) {
		this.classPackage = ClassDesc.getPackagePart(name);
		this.className = ClassDesc.getClassPart(name);
	}

	/**
	 * Set the package of this class.
	 */
	public void setPackage(String classPackage) {
		this.classPackage = classPackage;
	}

	/**
	 * Set the class name of this class' superclass.
	 */
	public void setParentName(String name) {
		this.parentName = name;
	}

	/**
	 * Set the package of this class' superclass.
	 */
	public void setParentPackage(String parentPackage) {
		this.parentPackage = parentPackage;
	}

	/**
	 * Set the full name of this class.
	 */
	public void setParentFullName(String name) {
		this.parentPackage = ClassDesc.getPackagePart(name);
		this.parentName = ClassDesc.getClassPart(name);
	}

	/**
	 * Return the package part of a full class name ("package.class").
	 */
	public static String getPackagePart(String desc) {
		int dot = desc.lastIndexOf('.');
		return (dot == -1 ? null : desc.substring(0, dot));
	}

	/**
	 * Return the class part of a package.class name.
	 */
	public static String getClassPart(String desc) {
		int dot = desc.lastIndexOf('.');
		return (dot == -1 ? desc : desc.substring(dot+1));
	}

	/**
	 * Return the name of this class, not including outer class(es).
	 */
	public String getInnerName() {
		int index = className.lastIndexOf('.');
		return (index == -1 ? className : className.substring(index+1));
	}

	/**
	 * Return the name of the parent class, not including outer
	 * class(es).
	 */
	public String getParentInnerName() {
		int index = parentName.lastIndexOf('.');
		return (index == -1 ? parentName : parentName.substring(index+1));
	}

	/**
	 * Return name of this class.
	 */
	public String getName() {
		return className;
	}

	/**
	 * Return the full name of this class.
	 */
	public String getFullName() {
		return (classPackage == null ? className : classPackage+"."+className);
	}

	/**
	 * Return the package name of this class.
	 */
	public String getPackage() {
		return classPackage;
	}

	/**
	 * Return the class name of this class' superclass.
	 */
	public String getParentName() {
		return parentName;
	}

	/**
	 * Return the package name of this class' superclass.
	 */
	public String getParentPackage() {
		return classPackage;
	}

	/**
	 * Return the full name of this class.
	 */
	public String getParentFullName() {
		return (parentPackage == null
				? parentName : parentPackage+"."+parentName);
	}

	/**
	 * Return number of constructors this class has.
	 */
	public int getConstructorCount() {
		return constructors.size();
	}

	/**
	 * Compare this class to another by comparing
	 * the class's full names.
	 */
	public int compareTo(Object o) {
		ClassDesc c = (ClassDesc) o;
		return this.getFullName().compareTo(c.getFullName());
	}

	/**
	 * Generate the Java source head for this class.
	 */
	public void generateHead(JavaSourceWriter out) {
		/* Print class header. */
		out.print(modifiers+" "+getInnerName()+" ");
		if (parentName != null
				&& parentPackage != null
				&& !(parentName.equals("Object")
					&& parentPackage.equals("java.lang"))) {
			out.print("extends "+parentName+" ");
		}

		/* Print implemented interfaces. */
		if (!implems.isEmpty()) {
			//out.println();
			out.print("implements ");
			Iterator it = implems.iterator();
			out.print(it.next());
			while (it.hasNext()) {
				out.print(", ");
				out.print(it.next());
			}
			out.print(" ");
		}

		/* Print class body. */
		out.println("{");
		out.increaseLevel();
	}

	/**
	 * Generate the Java source body for this class.
	 */
	public void generateBody(JavaSourceWriter out) {
		DeDocSource dedoc = (DeDocSource) DeDoc.getDefault();
		boolean first = true;
		boolean separate = dedoc.getSeparate();
		int bodyLevel = dedoc.getBodyLevel();

		/* Print fields. */
		for (Iterator it = fields.iterator() ; it.hasNext() ; ) {
			FieldDesc field = (FieldDesc) it.next();
			if (!first && separate) {
				out.println();
			}
			field.generate(out, bodyLevel);
			first = false;
		}

		/* Print constructors. */
		if (!separate && !first && !constructors.isEmpty()) {
			out.println();
			first = true;
		}
		for (Iterator it = constructors.iterator() ; it.hasNext() ; ) {
			ConstructorDesc constructor = (ConstructorDesc) it.next();
			if (!first && separate) {
				out.println();
			}
			constructor.generate(out, bodyLevel);
			first = false;
		}

		/* Print methods. */
		if (!separate && !first && !methods.isEmpty()) {
			out.println();
		}
		for (Iterator it = methods.iterator() ; it.hasNext() ; ) {
			MethodDesc method = (MethodDesc) it.next();
			if (!first && separate) {
				out.println();
			}
			method.generate(out, bodyLevel);
		}
		out.decreaseLevel();
		out.println("}");
	}

	/**
	 * Return a string representation of this class.
	 * This method returns the same as <tt>getFullName</tt>.
	 */
	public String toString() {
		return getFullName();
	}
}
