/* DeDoc.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.File;
import java.io.IOException;
import org.gjt.osk.util.Strings;
import org.gjt.osk.util.options.OptionParser;
import org.gjt.osk.util.options.OptionParserException;

/**
 * This is the superclass of all executable classes
 * in the DeDoc program.
 */
public abstract class DeDoc {
	/** The one any only live object of this class. */
	private static DeDoc defaultDeDoc;

	/** The version of DeDoc. */
	public static final String version = "0.13";

	/** The option parser used by DeDoc. */
	protected OptionParser op;

	/** The basename of this application */
	protected String baseName;

	/**
	 * Sole constructor.
	 */
	protected DeDoc() {
		defaultDeDoc = this;

		/* Initialize baseName. */
		String name = getClass().getName();
		int index = name.indexOf('.');
		baseName = (index == -1 ? name : name.substring(index+1));

		/* Initialize option parser. */
		op = new OptionParser();
		op.addOption("help");
		op.addOption("version");
		op.addOption("v|verbose");
	}

	/**
	 * Return the default instance of this class.
	 */
	public static DeDoc getDefault() {
		return defaultDeDoc;
	}

	/**
	 * Parse options and handle some default
	 * options.
	 */
	protected String[] parseOptions(String[] args) {
		String[] params = null;
		try {
			params = op.parse(args);
		} catch (OptionParserException e) {
			System.err.println(baseName+": "+e.getMessage());
			System.exit(1);
		}

		/* Check --version option. */
		if (op.isSelected("version")) {
			printVersion();
			System.exit(0);
		}

		/* Check --help option. */
		if (op.isSelected("help")) {
			printHelp();
			System.exit(0);
		}

		return params;
	}

	/**
	 * Return the value of a numeric option.
	 * Sometime in the future, this method will be
	 * incorporated into the <tt>OptionParser</tt>.
	 */
	protected int getNumericOptionValue(String name) {
		try {
			return Integer.parseInt(op.getValue(name));
		} catch (NumberFormatException e) {
			System.err.println(baseName+": --"+name
					+" option takes a numeric argument");
			System.exit(1);
			return -1;
		}
	}

	/**
	 * Print a message if in verbose mode.
	 */
	public void verbosePrint(String msg) {
		if (op.isSelected("v")) {
			System.err.println(msg);
		}
	}

	/**
	 * Called when a module has been completed.
	 */
	protected abstract void processModule(ModuleDesc module);

	/**
	 * Print a help message
	 */
	public void printHelp() {
		String[] helpHead = getHelpHeader();
		for (int c = 0 ; c < helpHead.length ; c++) {
			System.out.println(helpHead[c]);
		}

		System.out.println();
		System.out.println("Options:");

		String[] helpOptions = getHelpOptions();
		for (int c = 0 ; c < helpOptions.length ; c++) {
			System.out.println(helpOptions[c]);
		}

		String[] helpFooter = {
			"  -v, --verbose              explain what is being done",
			"      --help                 display this help and exit",
			"      --version              output version information and exit",
			"",
			"Report bugs to Oskar Liljeblad <osk@hem.passagen.se>.",
		};
		for(int c = 0 ; c < helpFooter.length ; c++) {
			System.out.println(helpFooter[c]);
		}
	}

	/**
	 * Return the help header.
	 */
	protected abstract String[] getHelpHeader();

	/**
	 * Return the additional help options.
	 */
	protected abstract String[] getHelpOptions();

	/**
	 * Print version information.
	 */
	public void printVersion() {
		System.out.println(baseName+" (DeDoc) "+version);
		System.out.println("Written by Oskar Liljeblad.");
		System.out.println();
		System.out.println("Copyright (C) 2000 Oskar Liljeblad.");
	}

	/**
	 * Report that an exception has been thrown.
	 */
	public void reportError(String msg, Exception e) {
		System.err.println(baseName+": "+msg+":");
		if (e instanceof IOException) {
			System.err.println(baseName+":   "+e.getMessage());
		} else {
			System.err.println(baseName+":   "+e);
		}
	}

	/**
	 * Process a file from an argument.
	 */
	protected void processArgument(String param) {
		File src = new File(param);
		String name = src.getName();

		if (src.isDirectory()) {
			File pkgFile = new File(src, "package-summary.html");
			File ovrFile = new File(src, "overview-summary.html");
			if (pkgFile.exists()) {
				JavaDocPackage docPackage = new JavaDocPackage();
				docPackage.read(pkgFile.getPath());
			} else if (ovrFile.exists()) {
				JavaDocOverview docOverview = new JavaDocOverview();
				docOverview.read(ovrFile.getPath());
			}
		} else if (name.equalsIgnoreCase("overview-summary.html")) {
			JavaDocOverview docOverview = new JavaDocOverview();
			docOverview.read(param);
		} else if (name.equalsIgnoreCase("package-summary.html")) {
			JavaDocPackage docPackage = new JavaDocPackage();
			docPackage.read(param);
		} else if (name.endsWith(".html")) {
			JavaDocClass docClass = new JavaDocClass();
			docClass.read(param);
		} else {
			System.err.println(baseName+": don't know what to do with '"
					+param+"'");
		}
	}
}
