/* DeDocDumper.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.Iterator;
import org.gjt.osk.util.options.OptionParser;

/**
 * DeDocDumper reads JavaDoc documentation and prints
 * Java source code that dumps all public final fields 
 * in the read documentation. This is useful when
 * implementing 
 */
public class DeDocDumper extends DeDoc {
	/** The file which generated code is written to. */
	private PrintWriter out;
	//private String outputFile;

	/**
	 * Constructor that parses the specified arguments.
	 */
	public DeDocDumper(String[] args) {
		super();

		op.addOption("o|output", OptionParser.REQUIRED_ARG);

		processOptions(args);
	}

	/**
	 * Process the options used by DeDocDumper.
	 */
	protected void processOptions(String[] args) {
		String[] params = super.parseOptions(args);

		/* Open output file. */
		String name = null;
		try {
			if (op.isSelected("output")) {
				name = op.getValue("output");
				out = new PrintWriter(new FileWriter(name), true);
			} else {
				name = "FieldDump";
				out = new PrintWriter(System.out, true);
			}
		} catch (IOException e) {
			reportError("could not open file '"+op.getValue("output")
					+"' for writing", e);
		}

		out.println("public class "+name+" {");
		out.println("\tpublic static void main(String[] args) {");

		/* Process each given file. */
		for (int c = 0 ; c < params.length ; c++) {
			processArgument(params[c]);
		}
		if (params.length == 0) {
			System.err.println(baseName+": nothing to do");
		}

		out.println("\t}");
		out.println("}");
	}

	/**
	 * Return the help header.
	 */
	protected String[] getHelpHeader() {
		return new String[] {
"Usage: "+baseName+" [OPTION]... SOURCE...",
"Read JavaDoc documents in SOURCE, print Java source code that dumps",
"all public final fields in the read documentation. SOURCE can be a ",
"directory, an 'overview-summary.html' file, a 'package-summary.html'",
"file, or another, HTML document containing JavaDoc documentation.",
		};
	}

	/**
	 * Return the additional help options.
	 */
	protected String[] getHelpOptions() {
		return new String[] {
"  -o, --output=FILE     where to place output (default is stdout)",
		};
	}

	/**
	 * Extract fields in the specified module.
	 */
	public void processModule(ModuleDesc module) {
		Iterator it = module.getClasses().iterator();
		while (it.hasNext()) {
			ClassDesc klass = (ClassDesc) it.next();
			Iterator it2 = klass.getFields().iterator();
			while (it2.hasNext()) {
				FieldDesc field = (FieldDesc) it2.next();
				if (field.getAccessModifier() == FieldDesc.PUBLIC
						&& field.hasModifier("final")
						&& field.hasModifier("static")
						&& field.isScalarType()) {
						
					String name = klass.getFullName()+"."+field.getName();
					out.println("\t\tSystem.out.println(\""+name+" = \"+"+name+");");
				}
			}
		}
	}

	/**
	 * Main method for this class. This method simply instantiates
	 * this class.
	 */
	public static void main(String[] args) {
		new DeDocDumper(args);
	}
}
