/* DeDocSource.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.File;
import org.gjt.osk.util.options.OptionParser;

/**
 * DeDocSource reads JavaDoc documentation in a directory,
 * and writes java source files into another directory. It
 * is run from the command line.
 *
 * <p>
 * NOTE: The usage documentation below may not be entirely up to
 * date. Please run DeDoc with the <tt>--help</tt> option instead.
 *
 * <p>
 * Usage:
 *
 * <p>
 * DeDoc [OPTION]... SOURCE...<br>
 * Read JavaDoc documents in SOURCE, and write Java source file(s)
 * into DEST (or current directory if not specified).
 *
 * <p>
 * Actions:
 * <ul>
 *
 * <li>If SOURCE is a directory with contains the file
 * overview-summary.html, or a file with the same name:
 * Generate Java source files for all packages and their classes in
 * the specified directory.
 * 
 * <li>If SOURCE is a directory with contains the file
 * package-summary.html, or a file with the same name:
 * Generate Java source files for all classes in the specified
 * directory.
 * 
 * <li>If SOURCE is another file with the extension .html:
 * Generate Java source files for the specified JavaDoc class
 * documentation file.
 * 
 * <li>If --version is supplied, print version information and exit.
 * 
 * <li>If --help is supplied, print help text and exit.
 * </ul>
 *
 * <p>
 * Options:
 * <ul>
 *
 * <li>-o, --output=DIRECTORY
 * <p>
 *   Specifies where to place output. This directory (and its parent
 *   directories) will be created if it does not exist.
 *
 * <li>-b, --body-level=LEVEL
 * <p>
 *   Specify what to generate for field initialization, constructor
 *   and method bodies. Possible values of LEVEL are:<br>
 *     0   Print no bodies at all. Print only semicolons after class
 *         members.<br>
 *     1   Print empty bodies, like "{ }" for all methods and
 *         constructors.<br>
 *     2   Print empty bodies. For methods that return a value,
 *         print return-statement to return default value according
 *         to method return type. This is the default. <br>
 *     3   Print code to generate debugging messages for each
 *         method and constructor.
 *
 * <li>--help
 * <p>
 * Display help and exit.
 *
 * <li>--version
 * <p>
 * Display version information and exit.
 *
 * <li>-s, --separate
 * <p>
 * Print a newline between each class member (methods and fields).
 *
 * </ul>
 */
public class DeDocSource extends DeDoc {
	/** Where to place generated source code. */
	private File outputDir;
	/** Value of the --body-level option. */
	private int bodyLevel;
	/** Value of the --wrap-width option. */
	private int wrapWidth;

	/**
	 * Constructor that parses the specified arguments.
	 */
	public DeDocSource(String[] args) {
		super();

		op.addOption("l|link", OptionParser.REQUIRED_ARG);
		op.addOption("j|junk-paths");
		op.addOption("o|output", OptionParser.REQUIRED_ARG);
		op.addOption("b|body-level", OptionParser.REQUIRED_ARG, "2");
		op.addOption("w|wrap-width", OptionParser.REQUIRED_ARG, "80");
		op.addOption("s|separate");

		processOptions(args);
	}

	/**
	 * Process the options used by DeDocSource.
	 */
	protected void processOptions(String[] args) {
		String[] params = super.parseOptions(args);
	
		/* Get body level value. */
		bodyLevel = getNumericOptionValue("body-level");
		if (bodyLevel < 0 || bodyLevel > 3) {
			System.err.println(baseName+": Body levels "
					+"0-3 are supported only.");
			System.exit(1);
		}

		/* Get wrapping width. */
		wrapWidth = getNumericOptionValue("wrap-width");

		/* Get destination directory. */
		if (op.isSelected("output")) {
			outputDir = new File(op.getValue("output"));
		}

		/* Process each given file. */
		for (int c = 0 ; c < params.length ; c++) {
			processArgument(params[c]);
		}
		if (params.length == 0) {
			System.err.println(baseName+": nothing to do");
		}
	}

	/**
	 * Return the help header.
	 */
	protected String[] getHelpHeader() {
		return new String[] {
"Usage: "+baseName+" [OPTION]... SOURCE...",
"Read JavaDoc documents in SOURCE, and write Java source file(s) into output",
"directory (or current directory if not specified). SOURCE can be a directory,",
"an 'overview-summary.html' file, a 'package-summary.html' file, or another",
"HTML document containing JavaDoc documentation.",
		};
	}

	/**
	 * Return the additional help options.
	 */
	protected String[] getHelpOptions() {
		return new String[] {
"  -o, --output=DIRECTORY     where to place output",
"  -b, --body-level=LEVEL     what to generate for method bodies (0-3)",
"  -l, --link=URL             link used when JavaDoc was generated",
"  -w, --wrap=WIDTH           word wrapping width (default is 80, 0 to disable)",
"  -j, --junk-paths           do not create directories for output",
"  -s, --separate             print empty line between class fields/methods",
		};
	}

	/**
	 * Return the value of the --body-level option.
	 */
	public int getBodyLevel() {
		return bodyLevel;
	}

	/**
	 * Return <tt>true</tt> if the --separate option is enabled.
	 */
	public boolean getSeparate() {
		return op.isSelected("s");
	}

	/**
	 * Return the value of the --link option, or <tt>null</tt>
	 * if it doesn't exist. The returned string will always
	 * have a trailing slashes.
	 */
	public String getHTMLBase() {
		String link = op.getValue("l");
		if (link == null) {
			return null;
		}
		return (link.endsWith("/") ? link : link+"/");
	}

	/**
	 * Return the word-wrapping width.
	 */
	public int getWrapWidth() {
		return wrapWidth;
	}

	/**
	 * Prepare the specified file for output.
	 * This involves creating the parent directories
	 * necessary.
	 */
	public String getOutputFile(String filename) {
		File file = new File(outputDir, filename);
		if (op.isSelected("j")) {
			file = new File(outputDir, file.getName());
		} else {
			file.getParentFile().mkdirs();
		}

		return file.getPath();
	}

	/**
	 * Write the finished module to a .java file.
	 */
	protected void processModule(ModuleDesc module) {
		verbosePrint("out (Java source) "+module.getOutputFile());
		module.write();
	}

	/**
	 * Main method for this class. This method simply instantiates
	 * this class.
	 */
	public static void main(String[] args) {
		new DeDocSource(args);
	}
}
