/* JavaDocClass.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.*;
import org.gjt.osk.util.Strings;
import org.gjt.osk.util.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.openxml.parser.HTMLParser;
import org.w3c.dom.*;

/**
 * JavaDocClass reads the HTML-documentation of a class or
 * interface generated by Sun Javadoc 1.2 (or later) and writes it
 * as a java source file.
 *
 * The body of methods and values of static fields in the generated
 * source file will be empty and uninitialized by default.
 * This program is however capable of writing debugging output code
 * like this:
 *
 * <pre>
 * public int method(int p1, String p2) {
 *     System.out.println("method(int p1="+p1+", String p2="+p2+")");
 *     return 0;
 * }
 * </pre>
 *
 */
public class JavaDocClass {
	private ModuleDesc module;
	private ClassDesc curClass;
	private String directory;
	private boolean write;

	/**
	 * Create a new JavaDoc class-reader that adds
	 * classes into a new <tt>ModuleDesc</tt> object.
	 */
	public JavaDocClass() {
		this.module = new ModuleDesc();
		this.write = true;
		curClass = new ClassDesc();
	}

	/**
	 * Create a new JavaDoc class-reader that adds
	 * classes into the specified <tt>ModuleDesc</tt> object.
	 */
	public JavaDocClass(ModuleDesc module) {
		this.module = module;
		this.write = false;
		curClass = new ClassDesc();
	}

	/**
	 * Return the current module description object.
	 */
	public ModuleDesc getModuleDesc() {
		return module;
	}

	/**
	 * Read and translate the classes in the specified
	 * HTML document into the current module.
	 */
	public void read(String htmlname) {
		DeDoc dedoc = DeDoc.getDefault();

		/* Remember the directory for the specified document. */
		directory = new File(htmlname).getParent();
		if (directory == null) {
			directory = ".";
		}

		/* Read and parse HTML document. */
		dedoc.verbosePrint("in (class JavaDoc) "+htmlname);
		try {
			FileReader reader = new FileReader(htmlname);
			InputSource input = new InputSource(reader);
			HTMLParser parser = new HTMLParser();
			parser.parse(input);
			translate(parser.getDocument());
		} catch (SAXException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		} catch (IOException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		}

		if (write) {
			dedoc.processModule(module);
		}
	}

	/**
	 * Translate a parsed HTML document.
	 */
	private void translate(Document document) {
		NodeState node = new NodeState(document);
		NodeState xNode;

		/* Skip stuff in the beginning. */
		node.seek("[#comment],HTML+HEAD,[SCRIPT],BODY+[#comment],[A],TABLE,"+
				  "[#comment],HR,[#comment],");
		xNode = node.seekTo("H2+FONT+");
		curClass.setPackage(xNode.getValue().trim());

		/* Skip "Direct Known Subclasses" section. */
		node.seek("H2,[PRE],[DL],[DL],[DL],HR,[B],[#text],[I],[P],DL");

		/* Skip past class documentation. */
		module.addClass(curClass);
		pickupClassModifiers(new NodeState(node));
		while (!node.trySeek("HR")) {
			node.seek(",");
		}
		node.seek(",");

		pickupInnerClasses(new NodeState(node));

		/* Process everything else. */
		String section = "";
		do {
			/* Identify current section. */
			if ((xNode = node.trySeekTo("TABLE+")) != null) {
				if (xNode.trySeek("TR+TD+FONT+B+#text")) {
					section = xNode.getValue();
				}
			}

			/* PRE-section: this is what's interesting.  */
			if ((xNode = node.trySeekTo("PRE+")) != null) {
				if (section.equals("Field Detail")) {
					processField(new NodeState(xNode));
				} else if (section.equals("Constructor Detail")) {
					processCallable(new NodeState(xNode), true);
				} else if (section.equals("Method Detail")) {
					processCallable(new NodeState(xNode), false);
				}
			}
		} while (node.trySeek(","));

		/* Add private constructor if current class is not an interface
		 * and has no constructors. This is the only private 'callable'
		 * that we automaticly create.
		 */
		if (!curClass.hasModifier("interface")
				&& curClass.getConstructorCount() == 0) {
			ConstructorDesc constructor = new ConstructorDesc(curClass);
			constructor.setModifiers("private");
			curClass.addConstructor(constructor);
		}
	}

	/**
	 * Merge a set of nodes into a string.
	 */
	private String mergeNodes(NodeState node) {
		StringBuffer data = new StringBuffer();

		/* Merge all nodes. */
		do {
			NodeState xNode;

			if ((xNode = node.trySeekTo("A+")) != null) {
				String klass = xNode.getValue();
				String link = node.getAttr("href");

				/* Import only outer class. */
				int pos = klass.indexOf('.');
				String impName = (pos == -1 ? klass : klass.substring(0, pos));
				module.addImportClass(getPackageFromPath(link)+"."+impName);

				data.append(klass);
				data.append(' '); //FIXME: should not be necessary, but is
			} else if ((xNode = node.trySeekTo("B+")) != null) {
				data.append(xNode.getValue());
				data.append(' ');
			} else if ((xNode = node.trySeekTo("I+")) != null) {
				data.append(xNode.getValue());
				data.append(' '); //FIXME: should not be necessary, but is
			} else if (node.trySeek("#text")) {
				String value = node.getValue(); //.trim();
				if (value.startsWith("[]") && data.toString().endsWith(" ")) {
					Strings.stripTrailingChars(data, " ");
				}
				data.append(value);
			}
		} while (node.trySeek(","));

		/* Clean up data. Remove concurrent whitespace. */
		Strings.translate(data, "\n\r\t\240", "    ");
		Strings.stripConcurrentChars(data, " ");
		return data.toString().trim();
	}

	/**
	 * Process a field description.
	 */
	private void processField(NodeState node) {
		String data = mergeNodes(node);
		int nameIndex = data.lastIndexOf(' ');
		FieldDesc field = new FieldDesc(curClass, data.substring(nameIndex+1));
		int typeIndex = data.lastIndexOf(' ', nameIndex-1);
		field.setType(data.substring(typeIndex+1, nameIndex));
		field.setModifiers(data.substring(0, typeIndex));
		curClass.addField(field);
	}

	/**
	 * Process a method/constructor description.
	 */
	private void processCallable(NodeState node, boolean constructor) {
		String data = mergeNodes(node);
		int paramsStart = data.indexOf('(');
		int paramsEnd = data.indexOf(')');
		int throwsStart = data.indexOf(" throws ");
		int modsEnd = paramsStart-1;
		while (data.charAt(modsEnd) == ' ') {
			modsEnd--;
		}
		modsEnd = data.lastIndexOf(' ', modsEnd);

		CallableDesc callable;
		if (constructor) {
			callable = new ConstructorDesc(curClass);
			callable.setModifiers(data.substring(0, modsEnd));
		} else {
			String name = data.substring(modsEnd, paramsStart).trim();
			MethodDesc method = new MethodDesc(curClass, name);
			callable = method;
			int typeStart = data.lastIndexOf(' ', modsEnd-1);
			method.setModifiers(data.substring(0, typeStart).trim());
			method.setReturnType(data.substring(typeStart+1, modsEnd).trim());
		}

		/* Handle parameters. */
		if (paramsStart+1 != paramsEnd) {
			int argumentCount = 0;
		
			String params = data.substring(paramsStart+1, paramsEnd).trim();
			StringTokenizer tok = new StringTokenizer(params, ",");
			while (tok.hasMoreTokens()) {
				String token = tok.nextToken().trim();
				int nameStart = token.indexOf(' ');
				String name;
				String param;
				if (nameStart != -1) {
					name = token.substring(nameStart+1);
					param = token.substring(0, nameStart);
				} else {
					name = "unknown_"+argumentCount++;
					param = token;
				}
				int arrayStart = name.indexOf(' ');
				if (arrayStart != -1) {
					param += name.substring(arrayStart+1);
					name = name.substring(0, arrayStart);
				}
				callable.addParameter(name, param);
			}
		}

		/* Handle exceptions. */
		if (throwsStart != -1) {
			String throwList = data.substring(throwsStart+8).trim();
			StringTokenizer tok = new StringTokenizer(throwList, ",");
			while (tok.hasMoreTokens()) {
				String token = tok.nextToken().trim();
				callable.addException(token);
			}
		}

		/* Now add the method/constructor to the class information class. */
		if (constructor) {
			curClass.addConstructor((ConstructorDesc) callable);
		} else {
			curClass.addMethod((MethodDesc) callable);
		}
	}

	/**
	 * Read inner classes section.
	 */
	private void pickupInnerClasses(NodeState node) {
		if (!node.trySeek("P")) {
			return;
		}

		NodeState xNode = node.seekTo("P+#comment");
		if (xNode.getValue().indexOf("= INNER CLASS SUMMARY =") == -1
				&& xNode.getValue().indexOf("= NESTED CLASS SUMMARY =") == -1) {
			System.err.println("Failed to find inner class summary!");
			return;
		}
		if (!xNode.trySeek(",A")) {
			return;
		}

		node.seek("P,TABLE+TR,");
		do {
			xNode = node.trySeekTo("TR+TD,");
			if (xNode == null) {
				return;
			}

			/* In "Inner classes inherited from ..." */
			if (!xNode.trySeek("TD+")) {
				return;
			}
			xNode.seek("CODE+B+A+#text");
			/*xNode.seek("CODE+");
			xNode.trySeek("B+");
			xNode.seek("A+#text");*/

			String name = directory+File.separator+xNode.getValue()+".html";
			JavaDocClass docClass = new JavaDocClass(module);
			docClass.read(name);
		} while (node.trySeek(",TR"));
	}

	/**
	 * Pickup class modifiers etc.
	 */
	private void pickupClassModifiers(NodeState node) {
		/* Pick up class modifiers. */
		node.seek("DL+DT+#text");
		curClass.setModifiers(node.getValue().trim());
		node.seek(",");

		/* Pick up class name. */
		NodeState xNode = node.seekTo("B+");
		curClass.setName(xNode.getValue().trim());
		node.seek("B,");

		/* Pick up parent class name and package. */
		if (node.trySeek("DT")) {
			node.seek("DT+#text");		// "extends"
			xNode = node;

			if (xNode.trySeek(",A")) {
				String name = getClassFromPath(xNode.getAttr("href"));
				String pkg = getPackageFromPath(xNode.getAttr("href"));
				curClass.setParentName(name.trim());
				curClass.setParentPackage(pkg.trim());

				/* Import only outer class. */
				int pos = name.indexOf('.');
				String impName = (pos == -1 ? name : name.substring(0, pos));
				module.addImportClass(pkg+"."+impName);
			} else {
				String full = xNode.getValue().trim();
				full = full.substring(8);		// skip past "extends "
				curClass.setParentFullName(full);

				/* XXX: We include inner class here too (but shouldn't),
				 * because there is no way of splitting package from class.
				 */
				module.addImportClass(full);
			}

			node.trySeek(",");
		}

		/* Deal with "implements" section. */
		if (node.trySeek("DT")) {
			node.seek("DT+#text");		// "implements"
			do {
				if ((xNode = node.trySeekTo("A+")) != null) {
					String link = node.getAttr("href");
					String desc = xNode.getValue();

					/* Import only outer class. */
					int pos = desc.indexOf('.');
					String imp = (pos == -1 ? desc : desc.substring(0, pos));
					module.addImportClass(getPackageFromPath(link)+"."+imp);

					curClass.addImplements(desc.trim());
				}
			} while (node.trySeek(",."));
		}
	}

	//////////////////////////////////////////////////////////////////
	//                                                              //
	//  Various helper methods                                      //
	//                                                              //
	//////////////////////////////////////////////////////////////////

	/**
	 * Extract the package part from a local path string.
	 */
	private String getPackageFromPath(String path) {
		DeDoc dedoc = DeDoc.getDefault();
		String link = (dedoc instanceof DeDocSource
				? ((DeDocSource) dedoc).getHTMLBase()
				: null);
		if (link != null && path.startsWith(link)) {
			path = path.substring(link.length());
		}

		/* Skip .html extension. */
		path = path.substring(0, path.length()-5);

		/* Just skip parent-directory references. */
		int pos;
		while ((pos = path.indexOf("../")) != -1) {
			path = path.substring(0, pos) + path.substring(pos+3);
		}

		pos	= path.lastIndexOf('/');
		return (pos == -1 ? null : path.substring(0, pos).replace('/','.'));
	}

	/**
	 * Extract the class part from a local path string.
	 */
	private String getClassFromPath(String path) {
		DeDoc dedoc = DeDoc.getDefault();
		String link = (dedoc instanceof DeDocSource
				? ((DeDocSource) dedoc).getHTMLBase()
				: null);
		if (link != null && path.startsWith(link)) {
			path = path.substring(link.length());
		}
	
		/* Skip .html extension. */
		path = path.substring(0, path.length()-5);

		/* Just skip parent-directory references. */
		int pos;
		while ((pos = path.indexOf("../")) != -1) {
			path = path.substring(0, pos) + path.substring(pos+3);
		}

		pos	= path.lastIndexOf('/');
		return (pos == -1 ? path : path.substring(pos+1).replace('/','.'));
	}
}
