/* JavaDocOverview.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.*;
import org.gjt.osk.util.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.openxml.parser.HTMLParser;
import org.w3c.dom.*;

/**
 * JavaDocOverview reads the HTML-documentation for an API, and
 * generates a set of <tt>ModuleDesc</tt> objects for all
 * classes in all packages documented.
 */
public class JavaDocOverview {
	private List packages = new LinkedList();
	private String directory;

	/**
	 * Construct a package reader.
	 */
	public JavaDocOverview() {
	}

	/**
	 * Return the list of packages read so far.
	 */
	public List getPackageNames() {
		return packages;
	}

	/**
	 * Read JavaDoc package documentation in the specified
	 * directory.
	 */
	public void read(String htmlname) {
		DeDoc dedoc = DeDoc.getDefault();

		/* Remember the directory for the specified document. */
		directory = new File(htmlname).getParent();
		if (directory == null) {
			directory = ".";
		}

		/* Read and parse HTML document. */
		dedoc.verbosePrint("in (overview JavaDoc) "+htmlname);
		try {
			FileReader reader = new FileReader(htmlname);
			InputSource input = new InputSource(reader);
			HTMLParser parser = new HTMLParser();
			parser.parse(input);
			translate(parser.getDocument());
		} catch (SAXException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		} catch (IOException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		}
	}

	/**
	 * Translate a parsed overview JavaDoc HTML document.
	 */
	private void translate(Document document) {
		NodeState node = new NodeState(document);

		/* Skip stuff in the beginning. */
		node.seek("[#comment],HTML+HEAD,[SCRIPT],BODY+[#comment],A,TABLE,"+
				  "[#comment],HR,[#comment],[CENTER],");
		node.trySeek("BODY+");	// Hack for badly formatted overviews

		/* Process each TABLE element found. */
		do {
			NodeState xNode;
		
			/* Identify current section. */
			if ((xNode = node.trySeekTo("TABLE+")) != null) {
				if (!xNode.trySeek("TR+TD+FONT+B+#text")) {
					continue;
				}

				xNode = node.seekTo("TABLE+TR");
				while (xNode.trySeek(",TR")) {
					processTableRow(new NodeState(xNode));
				}
			}
		} while (node.trySeek(","));
	}

	/**
	 * Process a row in a table with classes or interfaces.
	 */
	private void processTableRow(NodeState node) {
		node.seek("TR+TD+B+A");
		String file = node.getAttr("href");
		file = file.substring(0, file.lastIndexOf('.'));
		node.seek("+[I]+");

		/* Don't add inner static classes - that's done later. */
		String name = node.getValue();
		JavaDocPackage docPackage = new JavaDocPackage();
		docPackage.read(directory+File.separator+name.replace('.', '/')
				+File.separator+"package-summary.html");
		packages.add(name);
	}
}
