/* JavaDocPackage.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.*;
import org.gjt.osk.util.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.openxml.parser.HTMLParser;
import org.w3c.dom.*;

/**
 * JavaDocPackage reads the HTML-documentation for a complete
 * package, and generates a set of <tt>ModuleDesc</tt> objects.
 */
public class JavaDocPackage {
	private List modules;
	private String directory;

	/**
	 * Construct a package reader that keeps module
	 * descriptions in a new list.
	 */
	public JavaDocPackage() {
		this(new LinkedList());
	}

	/**
	 * Construct a package reader that keeps module
	 * descriptions in the specified list.
	 */
	public JavaDocPackage(List modules) {
		this.modules = modules;
	}

	/**
	 * Return the list of modules read so far.
	 */
	public List getModuleDescs() {
		return modules;
	}

	/**
	 * Read JavaDoc package documentation in the specified
	 * directory.
	 */
	public void read(String htmlname) {
		DeDoc dedoc = DeDoc.getDefault();

		/* Remember the directory for the specified document. */
		directory = new File(htmlname).getParent();
		if (directory == null) {
			directory = ".";
		}

		/* Read and parse HTML document. */
		dedoc.verbosePrint("in (package JavaDoc) "+htmlname);
		try {
			FileReader reader = new FileReader(htmlname);
			InputSource input = new InputSource(reader);
			HTMLParser parser = new HTMLParser();
			parser.parse(input);
			translate(parser.getDocument());
		} catch (SAXException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		} catch (IOException e) {
			dedoc.reportError("could not read "+htmlname, e);
			return;
		}
	}

	/**
	 * Translate a parsed package list JavaDoc HTML document.
	 */
	private void translate(Document document) {
		NodeState node = new NodeState(document);

		/* Skip stuff in the beginning. */
		node.seek("[#comment],HTML+HEAD,[SCRIPT],BODY+[#comment],A,TABLE,"+
				  "[#comment],HR,[#comment],H2");

		/* Process each TABLE element found. */
		while (node.trySeek(",")) {
			NodeState xNode;
		
			/* Identify current section. */
			if ((xNode = node.trySeekTo("TABLE+")) != null) {
				if (!xNode.trySeek("TR+TD+FONT+B+#text")) {
					continue;
				}

				xNode = node.seekTo("TABLE+TR");
				while (xNode.trySeek(",TR")) {
					processTableRow(new NodeState(xNode));
				}
			}
		}
	}

	/**
	 * Process a row in a table with classes or interfaces.
	 */
	private void processTableRow(NodeState node) {
		node.seek("TR+TD+B+A");
		String file = node.getAttr("href");
		file = file.substring(0, file.lastIndexOf('.'));
		node.seek("+[I]+");

		/* Don't add inner static classes - that's done later. */
		String name = node.getValue();
		if (name.indexOf('.') == -1) {
			JavaDocClass docClass = new JavaDocClass();
			docClass.read(directory + File.separator + name + ".html");
			modules.add(docClass.getModuleDesc());
		}
	}
}
