/* JavaSourceWriter.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;

/**
 * A class used to write a Java source file.
 */
public class JavaSourceWriter extends PrintWriter {
	private final int lineWidth;
	private String pending;
	private int indentLevel = 0;

	/**
	 * Construct a JavaSourceWriter that writes to the
	 * specified stream, with a default wrap width of 80.
	 */
	public JavaSourceWriter(File file)
	throws IOException {
		this(file, 80);
	}

	/**
	 * Construct a JavaSourceWriter that writes to the
	 * specified stream.
	 */
	public JavaSourceWriter(File file, int width)
	throws IOException {
		super(new FileOutputStream(file), true);
		this.lineWidth = (width <= 0 ? Integer.MAX_VALUE : width);
	}

	/**
	 * Print indentation.
	 */
	public void indent() {
		for (int c = 0 ; c < indentLevel ; c++) {
			super.print("\t");
		}
	}

	/**
	 * Print a line of new-line terminated text.
	 */
	private void printCompleteLine(String text) {
		StringBuffer line = new StringBuffer(text);
		if (pending != null) {
			line.append(pending);
			pending = null;
		}

		int textLen = line.length();
		boolean first = true;
		do {
			int base = 0;

			if (line.length() != 0) {
				if (first) {
					indent();
					base = indentLevel*4;
					first = false;
				} else {
					indentLevel += 2;
					indent();
					base = indentLevel*4;
					indentLevel -= 2;
				}
			}

			int subLen = Math.min(lineWidth-base, textLen);
			String subText = splitLine(line, subLen);
			textLen = line.length();
			super.print(subText);
			super.println();
		} while (textLen > 0);
	}

	/**
	 * Split line at specified position, removing it from the buffer
	 * and returning it.
	 */
	private String splitLine(StringBuffer line, int pos) {
		String delims = " (";
		int oldPos = pos;
		int len = line.length();

		/* Position is past buffer - remove everything from line. */
		if (pos >= len) {
			String text = line.toString();
			line.delete(0, pos);
			return text;
		}

		/* Find first word delimiter. */
		while (pos >= 0 && delims.indexOf(line.charAt(pos)) == -1) {
			pos--;
		}
		if (pos == -1) {
			pos = oldPos;
			while (pos < len && delims.indexOf(line.charAt(pos)) == -1) {
				pos++;
			}
		}

		/* Get word. */
		if (pos < len && line.charAt(pos) != ' ') {
			pos++;
		}
		String text = line.substring(0, pos);

		/* Skip past whitespace. */
		while (pos < len && line.charAt(pos) == ' ') {
			pos++;
		}

		line.delete(0, pos);
		return text;
	}

	/**
	 * Print a line of text that is not new-line terminated.
	 */
	public void print(String text) {
		pending = (pending == null ? text : pending+text);
	}

	/**
	 * Print a line of text that is not new-line terminated.
	 */
	public void print(Object o) {
		print(o.toString());
	}

	/**
	 * Print an empty line.
	 */
	public void println() {
		printCompleteLine("");
	}

	/**
	 * Increase the current indentation level.
	 */
	public void increaseLevel() {
		indentLevel++;
	}

	/**
	 * Decrease the current indentation level.
	 */
	public void decreaseLevel() {
		indentLevel--;
	}
}
