/* ModuleDesc.java
 *
 * Copyright (C) 2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.dedoc;

import java.io.*;
import java.util.*;

/**
 * Objects of this class represent Java source code files,
 * which may contain one or more classes.
 */
public class ModuleDesc {
	private SortedSet imports = new TreeSet();
	private SortedSet classes = new TreeSet();

	/**
	 * Create a new module.
	 */
	public ModuleDesc() {
	}

	/**
	 * Add an imported class.
	 */
	public void addImportClass(String cls) {
		/* No need to import classes in java.lang. */
		if (cls.startsWith("java.lang.") && cls.lastIndexOf('.') == 9) {
			return;
		}

		/* Should not include inner classes. */
		ClassDesc outermost = getOutermostClass();
		if (outermost != null && cls.startsWith(outermost.getFullName()+".")) {
			return;
		}

		/* No need to import the class itself. */
		if (outermost != null && outermost.getFullName().equals(cls)) {
			return;
		}

		imports.add(cls);
	}

	/**
	 * Add a class to this source file.
	 */
	public void addClass(ClassDesc classDesc) {
		classes.add(classDesc);
	}

	/**
	 * Remove unnecessary imports - that is,
	 * import statements for classes in this module.
	 */
	private void cleanImports() {
		Iterator it = classes.iterator();
		while (it.hasNext()) {
			ClassDesc classDesc = (ClassDesc) it.next();
			imports.remove(classDesc.getFullName());
		}
	}

	/**
	 * Return the outermost class.
	 */
	public ClassDesc getOutermostClass() {
		return (ClassDesc) classes.first();
	}

	/**
	 * Return the file that this module will write to.
	 */
	public String getOutputFile() {
		return ((DeDocSource) DeDoc.getDefault()).getOutputFile(
				getOutermostClass().getFullName()
				.replace('.', File.separatorChar)+".java");
	}

	/**
	 * Return the set of classes in the module.
	 */
	public Set getClasses() {
		return classes;
	}

	/**
	 * Generate a Java source file from current collected info
	 * to the specified file.
	 */
	public void write() {
		DeDocSource dedoc = (DeDocSource) DeDoc.getDefault();
		JavaSourceWriter out;

		String filename = getOutputFile();
		try {
			out = new JavaSourceWriter(new File(filename),
					dedoc.getWrapWidth());
		} catch (IOException e) {
			dedoc.reportError("could not write to "+filename, e);
			return;
		}

		/* Print package. */
		out.println("package "+getOutermostClass().getPackage()+";");
		out.println();

		/* Print imports. */
		boolean first = true;
		for (Iterator it = imports.iterator() ; it.hasNext() ; ) {
			String importDesc = (String) it.next();
			out.println("import "+importDesc+";");
			first = false;
		}

		/* Print classes. */
		if (!first && !classes.isEmpty()) {
			out.println();
			first = true;
		}

		Stack postponed = new Stack();
		for (Iterator it = classes.iterator() ; it.hasNext() ; ) {
			ClassDesc classDesc = (ClassDesc) it.next();

			/* Print class body of last class. */
			while (!postponed.empty()) {
				ClassDesc prevClass = (ClassDesc) postponed.peek();
				if (!classDesc.getName().startsWith(prevClass.getName())) {
					prevClass.generateBody(out);
					out.println();
					postponed.pop();
				} else {
					break;
				}
			}

			classDesc.generateHead(out);
			postponed.push(classDesc);
		}

		/* Print remaining bodies. */
		while (!postponed.empty()) {
			ClassDesc prevClass = (ClassDesc) postponed.pop();
			prevClass.generateBody(out);
			if (!postponed.empty()) {
				out.println();
			}
		}
	}
}
