/* Strings.java
 * 
 * Copyright (C) 1999-2000  Oskar Liljeblad
 *
 * This file is part of Oskar Liljeblad's Utility libraries.
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU Lesser General Public License. Please consult the file
 * "LICENSE" for details.
 */

package org.gjt.osk.util;

import java.lang.Character;	// necessary for JavaDoc links it seems?
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

/**
 * A helper-class with static methods for string parsing and manipulation.
 *
 * Possible todo:
 *   containsOnly  - true if a String contains only the listed characters
 *   containsAny   - true if a String contains any of listed characters
 *   replaceFirst  - with index, String/StringBuffer
 *   replace       - with index, String/StringBuffer
 *   replaceLast   - String/StringBuffer
 *   replaceLast   - with index, String/StringBuffer
 *
 * Definite todo:
 *   remove isJavaIdentifier
 *   rename isDigits to isDigitString, same for isWhitespaces and other
 */
public class Strings {
	/**
	 * Private constructor to indicate that class cannot be
	 * instantiated.
	 */
    private Strings() {
    }

	/**
	 * Uppercase only the first character in the specified
	 * string.
	 */
	public static String upperCaseFirst(String str) {
		return str.substring(0, 1).toUpperCase() + str.substring(1);
	}

	/**
	 * Lowercase only the first character in the specified
	 * string.
	 */
	public static String lowerCaseFirst(String str) {
		return str.substring(0, 1).toLowerCase() + str.substring(1);
	}

	/**
	 * Translate all ranges of whitespace characters in <tt>str</tt>
	 * into a single character (<tt>spaceChar</tt>).
	 */
	public static String stripConcurrentWhitespace(String str, char spaceChar) {
		StringBuffer buf = new StringBuffer(str.length());
		int strLen = str.length();
		boolean wasWS = false;

		for (int c = 0; c < strLen; c++) {
			char ch = str.charAt(c);
			boolean isWS = Character.isWhitespace(ch);

			if (isWS) {
				if (!wasWS)
					buf.append(spaceChar);
			} else {
				buf.append(ch);
			}
			wasWS = isWS;
		}

		return buf.toString();
	}

	/**
	 * Translate all ranges of whitespace characters in <tt>str</tt>
	 * into a single space.
	 */
	public static String stripConcurrentWhitespace(String str) {
		return stripConcurrentWhitespace(str);
	}

	/**
	 * Remove all concurrent occurencies of any of the listed
	 * characters in a string. For every character in the
	 * resulting string, if the character is in <tt>chars</tt>,
	 * the characters to the right or left may not also be in
	 * <tt>chars</tt>.
	 *
	 * <p>
	 * Examples:
	 * <table border=1>
	 * <tr>
	 * <td><b><tt>str</tt></b></td>
	 * <td><b><tt>chars</tt></b></td>
	 * <td><b>Result</b></td>
	 * </tr>
	 * <tr><td>aalpha</td><td>a</td><td>apha</td></tr>
	 * <tr><td>alaala</td><td>a</td><td>alala</td></tr>
	 * </table>
	 *
	 * <p>
	 * @return The stripped string.
	 * @param str The string to remove characters in.
	 * @param chars List of characters to remove.
	 */
	public static String stripConcurrentChars(String str, String chars) {
		StringBuffer strBuf = new StringBuffer(str);
		stripConcurrentChars(strBuf, chars);
		return strBuf.toString();
	}

	/**
	 * Remove all concurrent occurencies of any of the listed
	 * characters in a string buffer. For every character in the
	 * resulting string, if the character is in <tt>chars</tt>,
	 * the characters to the right or left may not also be in
	 * <tt>chars</tt>.
	 *
	 * @return The same string as <tt>str</tt>
	 * @param str The string buffer to remove characters in.
	 * @param chars List of characters to remove.
	 */
	public static StringBuffer stripConcurrentChars(StringBuffer str,
			String chars) {
		int len = str.length();
		int delStart = -1;

		for (int c = 0 ; c < len ; c++) {
			char ch = str.charAt(c);

			if (chars.indexOf(ch) != -1) {
				if (delStart == -1) {
					delStart = c;
				}
			} else {
				if (delStart != -1 && c != delStart+1) {
					str.delete(delStart+1, c);
					len -= c - (delStart+1);
					c = delStart+1;
				}
				delStart = -1;
			}
		}

		if (delStart != -1 && len != delStart+1) {
			str.delete(delStart+1, len);
		}

		return str;
	}


	/**
	 * Return the number of occurences of a character in the specified
	 * string.
	 *
	 * @return The number of occurences.
	 * @param str The string to look in.
	 * @param ch The character to look for.
	 */
	public static int countChar(String str, int ch) {
		int count = 0;
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			if (str.charAt(c) == ch)
				count++;
		}
		return count;
	}

	/**
	 * Return the number of occurences of any character in the
	 * specified string.
	 *
	 * @return The number of occurences.
	 * @param str The string to look in.
	 * @param chars The characters to look for.
	 */
	public static int countChars(String str, String chars) {
		int count = 0;
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			if (chars.indexOf(str.charAt(c)) != -1)
				count++;
		}
		return count;
	}

	/**
	 * Delete all occurencies of specified characters in a string
	 * buffer.
	 *
	 * @return The string buffer itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static StringBuffer deleteChars(StringBuffer str, String chars) {
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			if (chars.indexOf(str.charAt(c)) != -1)
				str.deleteCharAt(c);
		}
		return str;
	}

	/**
	 * Delete all occurencies of specified characters in a string.
	 *
	 * @return The string itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static String deleteChars(String str, String chars) {
		return deleteChars(new StringBuffer(str), chars).toString();
	}

	/**
	 * Delete all leading occurencies of specified characters in a
	 * string buffer.
	 *
	 * @return The string buffer itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static StringBuffer stripLeadingChars(StringBuffer str, String chars) {
		int len = str.length();
		int c = 0;
		for ( ; c < len && chars.indexOf(str.charAt(c)) != -1 ; c++);
		str.delete(0, c);
		return str;
	}

	/**
	 * Delete all leading occurencies of specified characters in a
	 * string.
	 *
	 * @return The string itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static String stripLeadingChars(String str, String chars) {
		int len = str.length();
		int c = 0;
		for ( ; c < len && chars.indexOf(str.charAt(c)) != -1 ; c++);
		return str.substring(c);
	}

	/**
	 * Delete all trailing occurencies of specified characters in a
	 * string buffer.
	 *
	 * @return The string buffer itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static StringBuffer stripTrailingChars(StringBuffer str, String chars) {
		int len = str.length();
		int c = len - 1;
		for ( ; c >= 0 && chars.indexOf(str.charAt(c)) != -1 ; c--);
		str.delete(c+1, len);
		return str;
	}

	/**
	 * Delete all trailing occurencies of specified characters in a
	 * string.
	 *
	 * @return The string itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static String stripTrailingChars(String str, String chars) {
		int len = str.length();
		int c = len - 1;
		for ( ; c >= 0 && chars.indexOf(str.charAt(c)) != -1 ; c--);
		return str.substring(0, c+1);
	}

	/**
	 * Delete all leading and trailing occurencies of specified
	 * characters in a string.
	 *
	 * @return The string itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static String stripChars(String str, String chars) {
		int len = str.length();

		int start = 0;
		for ( ;	start < len && chars.indexOf(str.charAt(start)) != -1
				; start++);

		int end = len - 1;
		for ( ; end >= start && chars.indexOf(str.charAt(end)) != -1 ; end--);

		return str.substring(start, end+1);
	}

	/**
	 * Delete all leading and trailing occurencies of specified
	 * characters in a string buffer.
	 *
	 * @return The string buffer itself
	 * @param str The string buffer to delete characters in.
	 * @param chars The characters to delete.
	 */
	public static StringBuffer stripChars(StringBuffer str, String chars) {
		stripLeadingChars(str, chars);
		stripTrailingChars(str, chars);
		return str;
	}

	/**
	 * Return the specified string repeated a number of times.
	 *
	 * @return The string with the repeated string.
	 * @param ch The string to repeat.
	 * @param count The number of times to repeat.
	 */
	public static String repeat(String str, int count) {
		StringBuffer res = new StringBuffer(count * str.length());
		for ( ; count > 0 ; count--) {
			res.append(str);
		}
		return res.toString();
	}

	/**
	 * Return the specified character repeated a number of times.
	 *
	 * @return The string with the repeated character.
	 * @param ch The character to repeat.
	 * @param count The number of times to repeat.
	 */
	public static String repeat(char ch, int count) {
		StringBuffer res = new StringBuffer(count);
		for ( ; count > 0 ; count--) {
			res.append(ch);
		}
		return res.toString();
	}

	/**
	 * Return the index of the first character in `chars' found in the
	 * specified string.
	 *
	 * @return An index in <tt>str</tt> or -1 if no character found.
	 * @param str The string to search in.
	 * @param chars The characters to look for.
	 */
	public static int indexOfAny(String str, String chars) {
		return indexOfAny(str, chars, 0);
	}

	/**
	 * Return the index of the first character in <tt>chars</tt> found
	 * in the specified string, starting at position <tt>start</tt>.
	 *
	 * @return An index in <tt>str</tt> or -1 if no character found.
	 * @param str The string to search in.
	 * @param chars The characters to look for.
	 * @param start Index in <tt>str</tt> to start look from.
	 */
	public static int indexOfAny(String str, String chars, int start) {
		int min = Integer.MAX_VALUE;
		int len = chars.length();
		for (int c = 0 ; c < len ; c++) {
			int pos = str.indexOf(chars.charAt(c), start);
			if (pos != -1 && pos < min) {
				min = pos;
			}
		}
		return (min == Integer.MAX_VALUE ? -1 : min);
	}

	/**
	 * Return the index of the first character in `chars' found in the
	 * specified string, moving backwards through the string.
	 *
	 * @return An index in <tt>str</tt> or -1 if no character found.
	 * @param str The string to search in.
	 * @param chars The characters to look for.
	 */
	public static int lastIndexOfAny(String str, String chars) {
		return lastIndexOfAny(str, chars, str.length()-1);
	}

	/**
	 * Return the last index of the first character in `chars' found in the
	 * specified string, moving backwards through the string.
	 *
	 * @return An index in <tt>str</tt> or -1 if no character found.
	 * @param str The string to search in.
	 * @param chars The characters to look for.
	 * @param start Index in <tt>str</tt> to start look at.
	 */
	public static int lastIndexOfAny(String s, String chars, int start) {
		int max = -1;
		int len = chars.length();
		for (int c = 0 ; c < len ; c++) {
			int pos = s.lastIndexOf(chars.charAt(c), start);
			if (pos != -1 && pos > max) {
				max = pos;
			}
		}
		return max;
	}

	/**
	 * Replace any occurences of a string in the specified string
	 * buffer.
	 *
	 * @return The string buffer itself.
	 * @param from The string to look for.
	 * @param to The string to replace with.
	 */
	public static StringBuffer replace(StringBuffer buf, String from,
			String to) {
		int pos = 0;
		int fromLen = from.length();

		while ((pos = buf.toString().indexOf(from, pos)) != -1) {
			buf.replace(pos, pos+fromLen, to);
			pos += fromLen;
		}

		return buf;
	}

	/**
	 * Replace any occurences of a string in the specified string.
	 *
	 * @return The string in which occurences of <tt>from</tt> are
	 *         replaced.
	 * @param from The string to look for.
	 * @param to The string to replace with.
	 */
	public static String replace(String str, String from, String to) {
		return replace(new StringBuffer(str), from, to).toString();
	}

	/**
	 * Replace the first occurence of a string in the specified string
	 * buffer.
	 *
	 * @return The string buffer itself.
	 * @param from The string to look for.
	 * @param to The string to replace with.
	 */
	public static StringBuffer replaceFirst(StringBuffer buf, String from,
			String to) {
		int pos = buf.toString().indexOf(from);
		if (pos != -1) {
			buf.replace(pos, pos+from.length(), to);
		}
		return buf;
	}

	/**
	 * Replace the first occurences of a string in the specified
	 * string.
	 *
	 * @return The string in which an occurence of <tt>from</tt> is
	 *         replaced.
	 * @param from The string to look for.
	 * @param to The string to replace with.
	 */
	public static String replaceFirst(String str, String from, String to) {
		int pos = str.indexOf(from);
		if (pos != -1) {
			return str.substring(0, pos)+to+str.substring(pos+from.length());
		}
		return str;
	}

	/**
	 * Translate characters in a string. Similar to the tr//
	 * operator in Perl.
	 *
	 * @return The translated string.
	 * @param str The string to translate.
	 * @param fromChars Characters to look for.
	 * @param toChars Characters to replace with.
	 */
	public static String translate(String str, String fromChars, String toChars) {
		char[] dest = str.toCharArray();
		for (int c = dest.length-1 ; c >= 0 ; c--) {
			int index = fromChars.indexOf(dest[c]);
			if (index != -1)
				dest[c] = toChars.charAt(index);
		}
		return new String(dest);
	}

	/**
	 * Translate characters in a string buffer. Similar to the tr//
	 * operator in Perl.
	 *
	 * @return The string buffer itself.
	 * @param str The string buffer to translate.
	 * @param fromChars Characters to look for.
	 * @param toChars Characters to replace with.
	 */
	public static StringBuffer translate(StringBuffer str, String fromChars, String toChars) {
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			int index = fromChars.indexOf(str.charAt(c));
			if (index != -1)
				str.setCharAt(c, toChars.charAt(index));
		}
		return str;
	}

	/**
	 * Return the specified collection as an array of strings by
	 * calling the <tt>toArray</tt> method.
	 *
	 * @return An array of strings.
	 * @param collection The collection to turn into strings.
	 */
	public static String[] toStringArray(Collection collection) {
		return (String[]) collection.toArray(new String[collection.size()]);
	}

	/**
	 * Split the specified string into tokens.
	 *
	 * @return An array of strings in the buffer.
	 * @param buf The string that will be tokenized.
	 * @param delim String of characters delimiting tokens.
	 */
	public static String[] tokenize(String buf, String delims) {
		int len = buf.length();
		List tokens = new LinkedList();

		int start = 0;
		while (true) {
			/* Skip leading delim characters. */
			while (start < len && delims.indexOf(buf.charAt(start)) != -1)
				start++;

			/* Find index of first delim-character. */
			int end = start;
			while (end < len && delims.indexOf(buf.charAt(end)) == -1)
				end++;

			/* Get word to extract. */
			String word = buf.substring(start, end);

			/* Skip trailing delim characters. */
			while (end < len && delims.indexOf(buf.charAt(end)) != -1)
				end++;

			start = end;
			if (word.length() == 0)
				break;
			else
				tokens.add(word);
		}

		return toStringArray(tokens);
	}

	/**
	 * Remove a token from the beginning of a <tt>StringBuffer</tt>
	 * and return it.
	 *
	 * @return The removed leading token (<tt>null</tt> if
	 *         <tt>buf</tt> is empty or consists of <tt>delim</tt>
	 *         characters only).
	 * @param buf The string that a word will be removed from.
	 * @param delims String of characters delimiting tokens.
	 */
	public static String pullToken(StringBuffer buf, String delims) {
		int len = buf.length();

		/* Skip leading delim characters. */
		int start = 0;
		while (start < len && delims.indexOf(buf.charAt(start)) != -1)
			start++;

		/* Find index of first delim-character. */
	/*	int end = indexOfAny(buf.toString(), delims, start);
		if (end == -1) {
			end = len;
		}*/
		int end = start;
		while (end < len && delims.indexOf(buf.charAt(end)) == -1)
			end++;

		/* Get word to extract. */
		String word = buf.substring(start, end);

		/* Skip trailing delim characters. */
		while (end < len && delims.indexOf(buf.charAt(end)) != -1)
			end++;

		/* Delete first token and delim characters. */
		buf.delete(0, end);

		return (word.length() == 0 ? null : word);
	}

	/**
	 * Remove a token from the end of a <tt>StringBuffer</tt>
	 * and return it.
	 *
	 * @return The removed trailing token (<tt>null</tt> if
	 *         <tt>buf</tt> is empty or consists of <tt>delim</tt>
	 *         characters only).
	 * @param buf The string that a word will be removed from.
	 * @param delim String of characters delimiting tokens.
	 */
	public static String pullLastToken(StringBuffer buf, String delim) {
		int len = buf.length();

		/* Skip trailing delim characters. */
		int end = len-1;
		while (end >= 0 && delim.indexOf(buf.charAt(end)) != -1) {
			end--;
		}

		/* Find index of last delim-character. */
		int start = lastIndexOfAny(buf.toString(), delim, end)+1;

		/* Get word to extract. */
		String word = buf.substring(start, end+1);

		/* Skip trailing delim characters. */
		do {
			start--;
		} while (start >= 0 && delim.indexOf(buf.charAt(start)) != -1);

		/* Delete first token and delim characters. */
		buf.delete(start+1, len);

		return (word.length() == 0 ? null : word);
	}

	/**
	 * Return <tt>true</tt> if the specified string consists of digits
	 * only. A character is considered a digit if
	 * {@link Character#isDigit(char)} returns <tt>true</tt>.
	 *
	 * @return <tt>true</tt> if <tt>str</tt> contains digits only.
	 * @param str The string to check.
	 */
	public static boolean isDigits(String str) {
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			if (!Character.isDigit(str.charAt(c))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Return <tt>true</tt> if the specified string consists of
	 * whitespaces only. A character is considered a whitespace if
	 * {@link Character#isWhitespace(char)} returns <tt>true</tt>.
	 *
	 * @return <tt>true</tt> if <tt>str</tt> contains whitespaces
	 *         only.
	 * @param str The string to check.
	 */
	public static boolean isWhitespaces(String str) {
		for (int c = str.length()-1 ; c >= 0 ; c--) {
			if (!Character.isWhitespace(str.charAt(c)))
				return false;
		}
		return true;
	}

	/**
	 * Return <tt>true</tt> if the specified string is a valid
	 * Java identifier (for example, a string which may be used
	 * as a variable name). Note that this method does not check if
	 * the specified string is a reserved keyword or not.
	 *
	 * @see java.lang.Character#isJavaIdentifierStart(char)
	 * @see	java.lang.Character#isJavaIdentifierPart(char)
	 */
	public static boolean isJavaIdentifier(String str) {
		int len = str.length();

		/* Check first character. */
		if (len < 1 || !Character.isJavaIdentifierStart(str.charAt(0)))
			return false;

		/* Check all other characters. */
		for (int c = 0 ; c < len ; c++) {
			if (!Character.isJavaIdentifierPart(str.charAt(c)))
				return false;
		}

		return true;
	}
}
