/* LookupMap.java
 * 
 * Copyright (C) 1999-2000  Oskar Liljeblad
 *
 * This file is part of Oskar Liljeblad's Utility libraries.
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU GPL License.  Please consult the file "COPYING" for details.
 */

package org.gjt.osk.util.adt;

import java.util.*;

/**
 * A class similar to the <tt>Map</tt>, but with the difference
 * that it is not necessary to know the length of the key
 * when looking up a value. This is good when you're searching
 * for a word (among a set of words) in a string.
 *
 * Todo:
 * - Use Object instead of String for keys?
 * - Speedup remove by counting length of all keys in the map.
 * - Extend AbstractMap, more support for Map.
 */
public class LookupMap /* extends AbstractMap */ {
	/** The map which stores keys and values. */
	private Map map;
	/** The length of the longest key in the map. */
	private int keyMaxLen = 0;
	/** The length of the shortest key in the map. */
	private int keyMinLen = Integer.MAX_VALUE;

	/** Construct a <tt>LookupMap</tt> backed by a <tt>HashMap</tt>. */
	public LookupMap() {
		this(new HashMap());
	}

	/** Construct a <tt>LookupMap</tt> backed by the specified map. */
	public LookupMap(Map map) {
		this.map = map;
	}

	/**
	 * Add a key-value association.
	 */
	public void put(String key, Object value) {
		map.put(key, value);
		keyMinLen = Math.min(keyMinLen, key.length());
		keyMaxLen = Math.max(keyMaxLen, key.length());
	}

	/**
	 * Return the object with the specified name (exact match).
	 *
	 * @see java.util.Map#get(java.lang.Object)
	 */
	public Object get(String key) {
		return map.get(key);
	}

	/**
	 * Return the key of the object that the specified string begins
	 * with, or <tt>null</tt> if none found.
	 */
	public String findLeading(String str) {
		for (int c = Math.min(keyMaxLen, str.length());
				c >= keyMinLen; c--) {
			String key = str.substring(0, c);

			if (map.containsKey(key)) {
				return key;
			}
		}
		return null;
	}

	/**
	 * Return the object which is associated with the key
	 * which the specified string starts with, or <tt>null</tt>
	 * if not found.
	 */
	public Object getLeading(String str) {
		String key = findLeading(str);
		return get(key);
	}

	/**
	 * Return the index of first key found in the specified string,
	 * or <tt>-1</tt> if none found.
	 *
	 * NOTE: untested when start != 0
	 */
	public int findFirst(String text, StringBuffer foundKeySlot, int start) {
		int end = text.length()-keyMinLen;

		for (int d = start ; d <= end ; d++) {
			for (int c = Math.min(d+keyMaxLen, end+keyMinLen) ; c >= d+keyMinLen ; c--) {
				String key = text.substring(d, c);

				if (map.containsKey(key)) {
					foundKeySlot.replace(0, foundKeySlot.length(), key);
					return d;
				}
			}
		}

		return -1;
	}

	/**
	 * Remove a key-value association.
	 */
	public void remove(String key) {
		map.remove(key);

		/* Update keyMinLen and keyMaxLen if necessary. */
		if (key.length() == keyMinLen || key.length() == keyMaxLen) {
			Iterator it = map.keySet().iterator();
			keyMaxLen = 0;
			keyMinLen = Integer.MAX_VALUE;
			while (it.hasNext()) {
				String thisKey = (String) it.next();
				int thisLen = thisKey.length();
				keyMinLen = Math.min(keyMinLen, thisLen);
				keyMaxLen = Math.max(keyMaxLen, thisLen);
			}
		}
	}
}
