/* NodeState.java
 * 
 * Copyright (C) 1999-2000  Oskar Liljeblad
 *
 * This file is part of DeDoc.
 *
 * This software is copyrighted work licensed under the terms of the
 * DeDoc License.  Please consult the file "LICENSE" for details.
 */

package org.gjt.osk.util.dom;

import java.util.Stack;
import java.util.StringTokenizer;
import org.w3c.dom.*;

/**
 * Represents a mutable or changable <tt>Node</tt> object,
 * while providing methods for traversing node trees and
 * accessing node attributes.
 */
public class NodeState {
	private Node node;
	private Stack nodeStack;

	/**
	 * Construct a node state object cloning another
	 * node state.
	 */
	public NodeState(NodeState nodeState) {
		this.node = nodeState.node;
		this.nodeStack = (Stack) nodeState.nodeStack.clone();
	}

	/**
	 * Construct a node state object specifying the
	 * initial node.
	 */
	public NodeState(Node node) {
		this.node = node;
		this.nodeStack = new Stack();
	}

	/**
	 * Construct a node state object specifying the
	 * initial node as a <tt>Document</tt>.
	 */
	public NodeState(Document document) {
		this(document.getFirstChild());
	}

	/**
	 * Return the node which this node state currently
	 * represents.
	 */
	public Node getNode() {
		return node;
	}

	/**
	 * Change the node which this node state represents.
	 */
	public void setNode(Node node) {
		this.node = node;
	}

	/**
	 * Return the name of this node, as returned by
	 * <tt>getNodeName</tt> in <tt>Node</tt>.
	 */
	public String getName() {
		return node.getNodeName();
	}

	/**
	 * Return the value of this node, as returned by
	 * <tt>getNodeValue</tt> in <tt>Node</tt>.
	 */
	public String getValue() {
		return node.getNodeValue();
	}

	/**
	 * Copy the current node into the node stack.
	 * It is left as the current node.
	 */
	public void push() {
		nodeStack.push(node);
	}

	/**
	 * Pop a node from the top of the stack, and make it the current.
	 * If the stack is empty, do nothing. (FIXME: Should we throw an
	 * exception?)
	 */
	public void pop() {
		if (!nodeStack.empty()) {
			node = (Node) nodeStack.pop();
		}
	}

	/**
	 * From this node, traverse tree according to specified element
	 * specification and return a <tt>NodeState</tt> for the node
	 * found. If a node was not found, throw a
	 * <tt>NodeStateException</tt>.
	 *
	 * <p>
	 * The element specification is a string containing one or more
	 * element names, separated by '<tt>,</tt>' (comma), '<tt>+</tt>'
	 * (plus), or '<tt>-</tt>' (minus).
	 *
	 * <p>
	 * A comma between two elements indicates that the elements
	 * should be siblings. A plus indicates that the second element
	 * should be a child of the previous one. A minus indicates
	 * that that the first element is a child of the second.
	 *
	 * <p>
	 * The element name '<tt>.</tt>' can be used to match an element
	 * of any name. If the node is enclosed in brackets (<tt>[]</tt>),
	 * its presence is optional.
	 *
	 * <p>
	 * @param elemSpec The element specification as described above.
	 * @return A new node state representing the found node.
	 * @throw NodeStateException If sought node could not be found.
	 */
	public NodeState seekTo(String elemSpec) {
		return new NodeState(seekNode(elemSpec));
	}

	/**
	 * Similar to <tt>seekTo</tt>, but changes the current
	 * <tt>NodeState</tt> object instead of creating a new one.
	 */ 
	public void seek(String elemSpec) {
		node = seekNode(elemSpec);
	}

	/**
	 * Similar to <tt>seekTo</tt>, but returns a <tt>Node</tt>
	 * instead of a newly allocated <tt>NodeState</tt> object.
	 */ 
	public Node seekToNode(String elemSpec) {
		return seekNode(elemSpec);
	}

	/**
	 * Similar to <tt>seekTo</tt>, but returns <tt>null</tt> if a new
	 * node was not found. This method will not throw a
	 * <tt>NodeStateException</tt>.
	 */ 
	public NodeState trySeekTo(String elemSpec) {
		try {
			return seekTo(elemSpec);
		} catch (NodeStateException e) {
			return null;
		}
	}

	/**
	 * Similar to <tt>seek</tt>, but returns a boolean indicating
	 * whether a new node was found (<tt>true</tt>) or not
	 * (<tt>false</tt>). This method will not throw a
	 * <tt>NodeStateException</tt>.
	 */ 
	public boolean trySeek(String elemSpec) {
		try {
			seek(elemSpec);
			return true;
		} catch (NodeStateException e) {
			return false;
		}
	}

	/**
	 * Similar to <tt>seekToNode</tt>, but returns a <tt>Node</tt>
	 * instead of a newly allocated <tt>NodeState</tt> object.
	 * If the specified element could not be found, return
	 * <tt>null</tt>.
	 */ 
	public Node trySeekToNode(String elemSpec) {
		try {
			return seekToNode(elemSpec);
		} catch (NodeStateException e) {
			return null;
		}
	}


	/**
	 * Find the attribute by the specified name and return its
	 * associated value. If the current node does not have the
	 * specified attribute, throw a <tt>NodeStateException</tt>.
	 */
	public String getAttr(String name) {
		NamedNodeMap map = node.getAttributes();
		if (map == null) {
			throw new NodeStateException(this, "attribute '"+name
					+"' not found");
		}
		Node attr = map.getNamedItem(name);
		if (attr == null) {
			throw new NodeStateException(this, "attribute '"+name
					+"' not found");
		}
		return attr.getNodeValue();
	}

	/**
	 * Similar to <tt>getAttr</tt> but doesn't throw any
	 * <tt>NodeStateException</tt>s. Instead, <tt>null</tt> is
	 * returned if the specified attribute could not be found.
	 */
	public String tryGetAttr(String name) {
		try {
			return getAttr(name);
		} catch (NodeStateException e) {
			return null;
		}
	}

	/**
	 * Return the same as <tt>toString</tt> in the node which
	 * this node state currently represents.
	 */
	public String toString() {
		return node.toString();
	}

	/**
	 * This method is used by the seek methods to find a node
	 * according to an element specification.
	 */
	private Node seekNode(String elemSpec) {
		StringTokenizer tok = new StringTokenizer(elemSpec, "+-,", true);
		Node curNode = node;
		boolean lastOptionalFailed = false;

		if (curNode == null) {
			throw new NodeStateException(this,
					"seek failed\ninitial node was null\n"+
					"when looking for '"+elemSpec+"'");
		}
		while (tok.hasMoreTokens()) {
			String token = tok.nextToken();

			if (token.equals("+")) {
				if (!lastOptionalFailed) {
					curNode = curNode.getFirstChild();
				}
				lastOptionalFailed = false;
			} else if (token.equals(",")) {
				if (!lastOptionalFailed) {
					curNode = curNode.getNextSibling();
				}
				lastOptionalFailed = false;
			} else if (token.equals("-")) {
				if (!lastOptionalFailed) {
					curNode = curNode.getParentNode();
				}
				lastOptionalFailed = false;
			} else if (token.startsWith("[") && token.endsWith("]")) {
				token = token.substring(1, token.length()-1);
				if (curNode == null) {
					lastOptionalFailed = true;
				}
				String name = curNode.getNodeName();
				if (!token.equals(".") && !name.equals(token)) {
					lastOptionalFailed = true;
				}
			} else {
				if (curNode == null) {
					throw new NodeStateException(this,
							"seek failed\nfound nothing, needed '"+
							token+"'\n"+"when looking for '"+elemSpec+"'");
				}
				String name = curNode.getNodeName();
				if (!token.equals(".") && !name.equals(token)) {
					throw new NodeStateException(this,
							"seek failed\nfound '"+name+"', needed '"+
							token+"'\n"+"when looking for '"+elemSpec+"'");
				}
			}
		}

		return curNode;
	}
}
