/* OptionParser.java
 * 
 * Copyright (C) 1999-2000  Oskar Liljeblad
 *
 * This file is part of Oskar Liljeblad's Utility libraries.
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU GPL License.  Please consult the file "COPYING" for details.
 */

package org.gjt.osk.util.options;

import java.util.*;
import org.gjt.osk.util.Strings;
import org.gjt.osk.util.adt.LookupMap;

/**
 * A parser for command line options. The interface has
 * been designed for simplicity. The parser is highly
 * configurable and can simulate POSIX getopt,
 * GNU long getopt and the parser of Sun's JDK tools.
 *
 * <p>
 * ChangeLog:
 * <ul>
 * <li><b>2000-08-04</b>:<br>
 * getNonOptions: new method:
 * LONG_ABBREV: Now implies LONG.
 * SHORT_BUNDLE: Now imples SHORT.
 * tokenize: Made a wee bit smaller.
 * processOption: New method provided for subclasses.
 * processNonOption: New method provided for subclasses.
 * parse: Now calls the process{Option,NonOption} methods.
 * Option: Now protected.
 * JavaOptionParser: New class.
 * </ul>
 *
 * <p>
 * Ideas/FIXME:
 * <ul>
 * <li>Support for reverting to default values for options. The method
 * <tt>clear</tt> would return options to default values (as set by
 * user). This requires two additional fields in Option:
 * <tt>defaultSelected</tt>, <tt>defaultValue</tt>. The method
 * <tt>addOption</tt> needs to be modified too.
 * <li>The method <tt>getNonOptions</tt> may need a better name,
 * as well as <tt>processNonOption</tt>.
 * <li>The <tt>OptionIterator</tt> class need to be rewritten,
 * removed entirely, or merged with this class. To merge, we need
 * to add: <tt>boolean hasNext()</tt>, <tt>String getNext()</tt>,
 * <tt>boolean isOption()</tt>, and maybe as a new method
 * <tt>isNextOption()</tt> too.
 * </ul>
 */
public class OptionParser {
	/** Type of option that takes no argument. */
	public static final int NO_ARG = 0;
	/** Type of option that takes a required argument. */
	public static final int REQUIRED_ARG = 1;
	/** Type for option that takes an optional argument. */
	public static final int OPTIONAL_ARG = 2;

	/** Flag indicating that prefix may introduce long options. */
	public static final int LONG = 1 << 0;
	/** Flag indicating that prefix may introduce short options. */
	public static final int SHORT = 1 << 1;
	/**
	 * Flag indicating that prefixed long options may be abbreviated.
	 * This implies the <tt>LONG</tt> flag.
	 */
	public static final int LONG_ABBREV = LONG | (1 << 2);
	/**
	 * Flag indicating prefixed short options may be bundled.
	 * This implies the <tt>SHORT</tt> flag.
	 */
	public static final int SHORT_BUNDLE = SHORT | (1 << 3);

	/* Complete state fields. */
	private SortedMap options = new TreeMap();
	private LookupMap prefices = new LookupMap();
	private String stopWord = "--";

	/* Parse state fields. */
	private List nonOptions = new LinkedList();
	/** Indicates whether the stop word has been read or not. */
	protected boolean noMoreOptions;
	private Option nextParamOption;
	private String nextParamOptionName;
	private String pendingShort;

	/**
	 * Class used to represent a single option
	 * with value.
	 */
	protected class Option {
		/**
		 * The names of the option, separated by pipe characters
		 * ('<tt>|</tt>')
		 */
		public String names;
		/** 
		 * The type of this option; <tt>NO_ARG</tt>,
		 * <tt>REQUIRED_ARG</tt> or <tt>OPTIONAL_ARG</tt>.
		 */
		public int type;
		/**
		 * True if this option was selected.
		 */
		public boolean selected;
		/**
		 * The current argument of this option, or <tt>null</tt> if
		 * the option doesn't take an argument.
		 */
		public String value;
	}

	/**
	 * Class used to represent a prefix introducing
	 * options on the command line.
	 */
	private class Prefix {
		public String introducer;
		public String delims;
		public boolean takeLong;
		public boolean takeShort;
		public boolean abbrevLong;
		public boolean bundleShort;
	}

	/**
	 * Create a new option parser. The default prefices simulates GNU
	 * <tt>getopt_long</tt> with a <tt>-W</tt> extension for long
	 * options.
	 */
	public OptionParser() {
		setPrefix("-W", LONG|LONG_ABBREV|SHORT|SHORT_BUNDLE, "=");
		setPrefix("--", LONG|LONG_ABBREV, "=");
		setPrefix("-", SHORT|SHORT_BUNDLE);
	}

	/**
	 * Set flags for an option prefix. Create it if necessary.
	 */
	public void setPrefix(String introducer, int flags, String delims) {
		Prefix prefix = new Prefix();
		prefix.introducer = introducer;
		prefix.takeLong = (flags & LONG) != 0;
		prefix.takeShort = (flags & SHORT) != 0;
		prefix.abbrevLong = (flags & LONG_ABBREV) != 0;
		prefix.bundleShort = (flags & SHORT_BUNDLE) != 0;
		prefix.delims = delims;

		prefices.put(introducer, prefix);
	}

	/**
	 * Set flags for an option prefix. Create it if necessary.
	 */
	public void setPrefix(String introducer, int flags) {
		setPrefix(introducer, flags, null);
	}

	/**
	 * Remove the prefix with the specified introducer.
	 */
	public void removePrefix(String introducer) {
		prefices.remove(introducer);
	}

	/**
	 * Add an option that takes no arguments
	 * (<tt>OptionParser.NO_ARG</tt>).
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names) {
		addOption(names, OptionParser.NO_ARG, false, null);
	}

	/**
	 * Add an option that takes no arguments
	 * (<tt>OptionParser.NO_ARG</tt>), with the specified
	 * default value.
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names, boolean selected) {
		addOption(names, OptionParser.NO_ARG, selected, null);
	}

	/**
	 * Add an option of the specified type -
	 * <tt>OptionParser.NO_ARG</tt>,
	 * <tt>OptionParser.REQUIRED_ARG</tt> or
	 * <tt>OptionParser.OPTIONAL_ARG</tt>.
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names, int type) {
		addOption(names, type, false, null);
	}

	/**
	 * Add an option that takes a required argument,
	 * with the specified default value.
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names, String value) {
		addOption(names, OptionParser.REQUIRED_ARG, false, value);
	}

	/**
	 * Add an option of the specified type,
	 * with the specified default value.
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names, int type, String value) {
		addOption(names, type, false, value);
	}

	/**
	 * Add an option of the specified type,
	 * with the specified default value.
	 * Option names are specified separated by the pipe character
	 * ('<tt>|</tt>').
	 */
	public void addOption(String names, int type, boolean selected) {
		addOption(names, type, selected, null);
	}

	/**
	 * Add an option internally.
	 */
	private void addOption(String names, int type, boolean selected,
			String value) {
		Option option = new Option();
		option.names = names;
		option.type = type;
		option.selected = selected;
		option.value = value;

		/* Split options, copy into Option structure. */
		StringTokenizer tokenizer = new StringTokenizer(names, "|");
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			options.put(token, option);
		}
	}

	/**
	 * Return value of a parsed option.
	 */
	public String getValue(String name) {
		Option option = (Option) options.get(name);
		return (option != null ? option.value : null);
	}

	/**
	 * Set the value of an option.
	 */
	public void setValue(String name, String value) {
		Option option = (Option) options.get(name);
		if (option != null) {
			option.value = value;
		}
	}

	/**
	 * Return <tt>true</tt> if option was found on the parsed
	 * lines.
	 */
	public boolean isSelected(String name) {
		Option option = (Option) options.get(name);
		return (option != null ? option.selected : false);
	}

	/**
	 * Set whether the option was found on the parsed lines.
	 */
	public void setSelected(String name, boolean selected) {
		Option option = (Option) options.get(name);
		if (option != null) {
			option.selected = selected;
		}
	}

	/**
	 * Clear the state of this option parser.
	 * Reset all values of option to <tt>null</tt>.
	 */
	public void clear() {
		/* FIXME: This method is unnecessarily slow because
		 * the same option is cleared multiple times if it
		 * has multiple names.
		 */
		Iterator it = options.values().iterator();
		while (it.hasNext()) {
			Option option = (Option) it.next();
			option.selected = false;
			option.value = null;
		}
	}

	/** 
	 * Return an option iterator for the specified array of
	 * strings.
	 */
	public OptionIterator tokenize(String[] args) {
		return new OptionIterator(this, args);
	}

	/**
	 * Parse the specified array of strings.
	 * An array of non-option strings is returned.
	 */
	public String[] parse(String[] args)
	throws OptionParserException {
		nonOptions.clear();
		noMoreOptions = false;
		nextParamOption = null;
		nextParamOptionName = null;

		for (int c = 0 ; c < args.length ; c++) {
			parse(args[c], null);
		}
		if (nextParamOption != null) {
			throw new MissingParameterException(nextParamOptionName);
		}
		return getNonOptions();
	}

	/**
	 * Return the array of non-option arguments found so far.
	 */
	public String[] getNonOptions() {
		return (String[])
				nonOptions.toArray(new String[nonOptions.size()]);
	}

	/**
	 * Called internally by the <tt>OptionIterator</tt> to
	 * parse the next argument.
	 */
	void parseNext(OptionIterator oi)
	throws OptionParserException {
		while (pendingShort != null || oi.position < oi.args.length) {
			boolean result;
			if (pendingShort != null) {
				result = parse(pendingShort, oi);
			} else {
				result = parse(oi.args[oi.position], oi);
				oi.position++;
			}

			if (result) {
				break;
			}
		}


		if (nextParamOption != null) {
			throw new MissingParameterException(nextParamOptionName);
		}
		oi.completed = (pendingShort == null && oi.position >= oi.args.length);
	}

	/**
	 * Simple help method to update the option iterator's state.
	 *
	 * FIXME: make some guidelines for the value of <tt>value</tt> when
	 * <tt>isOption</tt> is true.
	 */
	private void updateOptionIterator(OptionIterator oi, boolean isOption,
			String value) {
		if (oi != null) {
			oi.nextValue = value;
			oi.nextIsOption = isOption;
			pendingShort = null;//FIXME
		}
	}

	/**
	 * Parse a single string part of a string array.
	 *
	 * @return <tt>true</tt> if a complete option/non-option was
	 *         not retreived.
	 */
	private boolean parse(String arg, OptionIterator oi)
	throws OptionParserException {
		Prefix prefix;

		/* Check if we are waiting for an option argument. */
		if (nextParamOption != null) {
			updateOptionIterator(oi, true, getFirstName(nextParamOption.names));
			nextParamOption.value = arg;
			nextParamOption.selected = true;
			processOption(nextParamOptionName, nextParamOption);
			nextParamOption = null;
			return true;
		}

		/* Already finished? */
		if (noMoreOptions) {
			nonOptions.add(arg);
			processNonOption(arg);
			updateOptionIterator(oi, false, arg);
			return true;
		}

		/* Non-option introducer? */
		if (stopWord != null && arg.equals(stopWord)) {
			noMoreOptions = true;
			return false;
		}

		/* Remove prefix. If no prefix, return as non-option. */
		prefix = getStringPrefix(arg);
		if (prefix == null || arg.equals(prefix.introducer)) {
			nonOptions.add(arg);
			processNonOption(arg);
			updateOptionIterator(oi, false, arg);
			return true;
		}
		arg = arg.substring(prefix.introducer.length());

		/* Manage long options */
		if (prefix.takeLong) {
			String optArg = getStringArgument(prefix, arg);
			String optName = (optArg == null ? arg
					: arg.substring(0, arg.length() - optArg.length() - 1));
				/* Look for exact match of option. */
			if (prefix.takeShort || optName.length() != 1) {
				Option option = (Option) options.get(optName);
					/* Got complete, unabbreviated long/short option? */
				if (option != null) {
					if (option.type == OptionParser.REQUIRED_ARG
							&& optArg == null) {
						nextParamOption = option;
						nextParamOptionName = optName;
						return false;
					}
					if (option.type == OptionParser.NO_ARG
							&& optArg != null) {
						updateOptionIterator(oi, false, null);
						throw new UnexpectedParameterException(optName);
					}
					option.value = optArg;
					option.selected = true;
					updateOptionIterator(oi, true, getFirstName(option.names));
					processOption(optName, option);
					return true;
				}
			}

			/* Maybe an abbreviated long option? */
			if (prefix.abbrevLong) {
				/* Special case - if takeLong & takeShort & longAbbrev,
				 * before checking for a long abbreviation, check for
				 * the possibility of short options instead.
				 */
				if (!prefix.takeShort || !isGoodShort(arg, prefix)) {
					Option option = getUniqueOption(optName, prefix);
					/* Ambiguous or unabbreviated option?
					 * (getUniqueOption will not return null.
					 * Instead, it will throw an exception
					 * which propagates through this method.)
					 */

					/* Got abbreviated long option */
					if (option.type == OptionParser.REQUIRED_ARG
							&& optArg == null) {
						nextParamOption = option;
						nextParamOptionName = optName;
						return false;
					}
					if (option.type == OptionParser.NO_ARG
							&& optArg != null) {
						updateOptionIterator(oi, false, null);
						throw new UnexpectedParameterException(optName);
					}
					option.value = optArg;
					option.selected = true;
					updateOptionIterator(oi, true, getFirstName(option.names));
					processOption(optName, option);
					return true;
				}
			}

			/* Only alternative left - short option */
			if (!prefix.takeShort) {
				updateOptionIterator(oi, false, null);
				throw new UnrecognizedOptionException(optName);
			}

			/* In a few cases, this line may be reached.
			 * That is when a long option was not matched,
			 * but may be a short option (or bundled short options).
			 */
		}

		/* Prefix does not introduce long or short options?
		 * It isn't much useful then.
		 */
		if (!prefix.takeShort) {
			/* The programmer most likely made an error.
			 * Here we will treat the parameter as a non-option.
			 */
			arg = prefix.introducer + arg;
			nonOptions.add(arg);
			updateOptionIterator(oi, false, arg);
			processNonOption(arg);
			return true;
		}

		/* Manage short (usually bundled) options. */
		while (arg.length() > 0) {
			Option option = (Option) options.get(arg.substring(0,1));

			/* Option really exists? */
			if (option == null) {
				updateOptionIterator(oi, false, null);
				throw new UnrecognizedOptionException(arg.substring(0,1));
			}

			/* Got a single option that take parameter? */
			if (option.type == OptionParser.REQUIRED_ARG) {
				if (arg.length() == 1) {
					nextParamOption = option;
					nextParamOptionName = arg.substring(0,1);
					updateOptionIterator(oi, false, null);
					return false;
				}
				option.value = arg.substring(1);
				option.selected = true;

				/* Abort the loop - everything has been parsed. */
				updateOptionIterator(oi, true, getFirstName(option.names));
				processOption(arg.substring(0, 1), option);
				return true;
			}

			if (option.type == OptionParser.OPTIONAL_ARG) {
				option.value = (arg.length() == 1 ? null : arg.substring(1));
				option.selected = true;

				/* Abort the loop - everything has been parsed. */
				updateOptionIterator(oi, true, getFirstName(option.names));
				processOption(arg.substring(0, 1), option);
				return true;
			}

			/* Got more than one short option, but bundling disallowed? */
			if (!prefix.bundleShort && arg.length() != 1) {
				updateOptionIterator(oi, false, null);
				throw new UnrecognizedOptionException(arg);
			}

			option.selected = true;
			arg = arg.substring(1);

			if (oi != null) {
				updateOptionIterator(oi, true, getFirstName(option.names));
				pendingShort = (arg.length() == 0 ? null : prefix.introducer+arg);
				return true;
			}
		}

		/* This can't (at least shouldn't) be reached when oi is not null. */
		return true;
	}

	/**
	 * Return the first short name in the option name
	 * specifiation string. This method is provided for
	 * convenience - it can be useful when printing help
	 * messages from custom option information arrays.
	 */
	public static String getFirstShortName(String names) {
		StringTokenizer tokenizer = new StringTokenizer(names, "|");
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			if (token.length() == 1) {
				return token;
			}
		}
		return null;
	}

	/**
	 * Return the first long name in the option name
	 * specifiation string. This method is provided for
	 * convenience - it can be useful when printing help
	 * messages from custom option information arrays.
	 */
	public static String getFirstLongName(String names) {
		StringTokenizer tokenizer = new StringTokenizer(names, "|");
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			if (token.length() > 1) {
				return token;
			}
		}
		return null;
	}

	/**
	 * Return the first name in the option name
	 * specifiation string. This method is provided for
	 * convenience - it can be useful when printing help
	 * messages from custom option information arrays.
	 */
	public static String getFirstName(String names) {
		int index = names.indexOf('|');
		return (index == -1 ? names : names.substring(0, index));
	}


	/**
	 * Set the string that ends options and introduces
	 * non-options. This can be set to <tt>null</tt> to
	 * disable the stop word.
	 */
	public void setStopWord(String str) {
		stopWord = str;
	}

	/**
	 * Return the complete name of an abbreviated option
	 * if abbreviated uniquely.
	 *
	 * The second argument indicates whether a short
	 * option may be returned if it may be ambiguous.
	 *
	 * @returns <tt>null</tt> if there are more than one option
	 * that the given option could be an abbreviation of.
	 * Also returns <tt>null</tt> is there is no candidate at all.
	 */
	private Option getUniqueOption(String asked, Prefix prefix)
	throws OptionParserException {
		boolean reqLong = !prefix.takeShort;

		/* If there is an exact match, then we consider the
		 * asked option to be unique.
		 */
		Option option = (Option) options.get(asked);
		if (option != null && !(reqLong && asked.length() == 1)) {
			return option;
		}

		/* Get matches. */
		SortedMap results = options.subMap(asked, asked+"\uFFFF");
		int size = results.size();
		if (size == 0) {
			throw new UnrecognizedOptionException(asked);
		}

		/* Skip first found if short option and long required. */
		String name = (String) results.firstKey();
		if (reqLong && name.length() == 1) {
			size--;
			if (size == 0) {
				throw new UnrecognizedOptionException(asked);
			}
		}

		/* The found option is unique if there is only one alternative. */
		if (size == 1) {
			return (Option) options.get(results.lastKey());
		}

		/* Check if the alternatives are unique or the same. */
		Iterator it = results.keySet().iterator();
		if (reqLong && name.length() == 1) {
			it.next();
		}

		option = (Option) options.get((String) it.next());
		while (it.hasNext()) {
			name = (String) it.next();
			Option candidate = (Option) options.get(name);
			if (option != candidate) {
				throw new AmbiguousOptionException(asked);
			}
		}

		return option;
	}

	/**
	 * Return <tt>true</tt> if string is a correct set of short
	 * options.
	 */
	private boolean isGoodShort(String str, Prefix prefix) {
		for (int c = 0 ; c < str.length() ; c ++) {
			Option option = (Option) options.get(str.substring(c,c+1));

			/* Option doesn't even exist? */
			if (option == null) {
				return false;
			}
			/* Last option char or option takes argument? */
			if (c == str.length()-1
					|| option.type != OptionParser.NO_ARG) {
				break;
			}
			/* More than one char, but no bundling? */
			if (!prefix.bundleShort) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Return the argument of a string of an option with
	 * included value (--option=value).
	 */
	private String getStringArgument(Prefix prefix, String str) {
		if (prefix.delims != null) {
			int pos = Strings.indexOfAny(str, prefix.delims);
			if (pos != -1) {
				return str.substring(pos+1);
			}
		}
		return null;
	}

	/**
	 * Return the prefix of a string, or <tt>null</tt> if none.
	 */
	private Prefix getStringPrefix(String asked) {
		return (Prefix) prefices.getLeading(asked);
	}

	/**
	 * This method is called when an option has just been modified
	 * (selected, or has had its value set). It is provided for
	 * subclasses, and does nothing by default. Note that this method
	 * is called after a possible change of value or selected status
	 * has been done.
	 *
	 * @param The exact name (short or long) of the parsed option.
	 * @param The option structure.
	 */
	protected void processOption(String name, Option option) {
	}

	/**
	 * This method is called when a non-option argument was just
	 * found. It is provided for subclasses, and does nothing by
	 * default.
	 */
	protected void processNonOption(String text) {
	}
}
